from abaqusGui import *
from abaqusConstants import *
import copy

from abq_Pme.pmeConstants import *
from abq_Pme.pmeTextFieldDB import PmeTextFieldDB
from abq_Pme.pmeComboBoxDB import PmeComboBoxDB
from abq_Pme.pmeVisibleColumnsDB import PmeVisibleColumnsDB

meshIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconMesh.png'))
crossIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconCross.png'))
deactivatedIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconDeactivated.png'))
shellFreeIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconShellFree.png'))
shellStructuredIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconShellStructured.png'))
shellSweepIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconShellStructured.png'))
solidFreeIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidFree.png'))
solidSweptIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidSwept.png'))
solidStructuredIcon =  afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidStructured.png'))

deactivatedIcon.create()
meshIcon.create()
crossIcon.create()
shellFreeIcon.create()
shellStructuredIcon.create()
shellSweepIcon.create()
solidFreeIcon.create()
solidSweptIcon.create()
solidStructuredIcon.create()
    
def getReverse(d):
    rev = {}
    for key,value in d.items():
        rev[value] = key
    return rev


defaultColIdToColTypeMap = {0:STATUS,
                            1:NAME,
                            2:TECHNIQUE,
                            3:ELEMSHAPE,
                            4:GENERATIONMETHOD,
                            5:GEOMETRICORDER,
                            6:FORMULATION,
                            7:ELEMSIZE,
                            8:MATERIAL}
# tableFeatures = (columnType,     title,                   entryType,      colWidth, columnJustify,     isEditable)
tableFeatures = {STATUS:           ('Status',                AFXTable.TEXT,  75,       AFXTable.CENTER,   False),
                 NAME:             ('Part Name',             AFXTable.TEXT,  150,      AFXTable.LEFT,     False),
                 TECHNIQUE:        ('Technique',             AFXTable.TEXT,  90,       AFXTable.CENTER,   False),
                 ELEMSHAPE:        ('Element\nShape',        AFXTable.LIST,  100,      AFXTable.LEFT,     True),
                 GENERATIONMETHOD: ('Method',                AFXTable.LIST,  100,      AFXTable.LEFT,     True),
                 GEOMETRICORDER:   ('Geometric\nOrder',      AFXTable.LIST,  85,       AFXTable.LEFT,     True),
                 FORMULATION:      ('Formulation',           AFXTable.LIST,  165,      AFXTable.LEFT,     True),
                 ELEMSIZE:         ('Element\nSize',         AFXTable.FLOAT, 75,       AFXTable.LEFT,     True),
                 MATERIAL:         ('Material',              AFXTable.LIST,  150,      AFXTable.LEFT,     True),
                 ERRORS:           ("Elements\nErrors (%)",  AFXTable.FLOAT, 80,       AFXTable.LEFT,     False),
                 WARNINGS:         ("Elements\nWarnings (%)",AFXTable.FLOAT, 95,       AFXTable.LEFT,     False),
                 RIGID:            ("Rigid",                 AFXTable.BOOL,  60,       AFXTable.CENTER,   True),
                 SKIN:             ("Skin",                  AFXTable.BOOL,  60,       AFXTable.CENTER,   True),
                 THICKNESS:        ("Thickness",             AFXTable.FLOAT, 85,       AFXTable.LEFT,     True),
                 VOLUME:           ("Volume",                AFXTable.FLOAT, 65,       AFXTable.LEFT,     False),
                }
    
###########################################################################
# Class definition
###########################################################################

class PmeModelDB(AFXDataDialog):
    [
        ID_TABLE,
        ID_MESH,
        ID_SKIP,
        ID_HEX,
        ID_QUAD_SHELL,
        ID_TET,
        ID_DIRECT,
        ID_VIRTUAL_TOPOLOGY,
        ID_EXTRUDE,
        ID_AUTOMATIC,
        ID_LINEAR,
        ID_QUADRATIC,
        ID_RIGID_ON,
        ID_RIGID_OFF,
        ID_SKIN_ON,
        ID_SKIN_OFF,
        ID_SKIN_PICK_SURFACE,
        ID_EDIT_VISIBLE,
        ID_REDUCED_INTEGRATION,
        ID_FULL_INTEGRATION,
        ID_INCOMPATIBLE,
        ID_CONTINUUM_SHELL,
        ID_MODIFIED,
        ID_IMPROVED,
        ID_REDUCED_INTEGRATION_HYBRID,
        ID_FULL_INTEGRATION_HYBRID,
        ID_INCOMPATIBLE_HYBRID,
        ID_MODIFIED_HYBRID,
        ID_RESUME,
        ID_SUPPRESS,
        ID_PICK_SURFACE,
        ID_ERRORS,
        ID_WARNINGS,
        ID_LAST,
        ] = range(AFXDataDialog.ID_LAST, AFXDataDialog.ID_LAST+34)
    numInstance = 0
    menuPaneItr = 0
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, form):
    
        PmeModelDB.numInstance += 1
        PmeModelDB.menuPaneItr += 1

        AFXDataDialog.__init__(self, form, 'Part Manager Extension', self.OK|self.APPLY|self.DISMISS,
                               DECOR_RESIZE|DIALOG_ACTIONS_SEPARATOR|DATADIALOG_BAILOUT)


        #GUILOG
        getAFXAliasMap().setPrefix(self, 'Pme-%d' % PmeModelDB.numInstance)
        #ENDGUILOG

        self.form = form
        self.sortColumn = 1
        self.sortOrder = -1
        self.currentColIdToColTypeMap = copy.deepcopy(defaultColIdToColTypeMap) # {columnId:columnType}
        self.currentColTypeToColIdMap = getReverse(self.currentColIdToColTypeMap) # {columnType:columnID}
        self.partNameToRowMap = {} # {partName:rowId}
        self.rowToPartNameMap = {} # {rowId:partName}
        self.hiddenTableMap = {} # {partName:{columnType:value})
        self.newVolumeMap = {} # {partName:Volume}

        self.textDB = None

        # Element Shapes
        self.elemShapeMap = {} # = {technqiue:(elemShape1, elemShape2,...)} default is elemShape1
        self.elemShapeListIdMap = {} # = {technqiue:listId}
        self.elemShapeMap["Solid Sweep"] =  ("Hex", "Tet", "Quad Shell")
        self.elemShapeMap["Solid Structured"] =  ("Hex", "Tet", "Quad Shell")
        self.elemShapeMap["Solid Free"] = ("Hex", "Tet", "Quad Shell")
        self.elemShapeMap["Shell Free"] = ("Quad Shell", )
        self.elemShapeMap["Shell Sweep"] = ("Quad Shell", )
        self.elemShapeMap["Shell Structured"] = ("Quad Shell", )

        # Generation Methods
        self.generationMethodMap = {} # map = {technique:{elemShape:(method1, method2, ...)}}
        self.generationMethodListIdMap = {} # map = {technique:{elemShape:(lisId1, listId2, ...)}}
        for technique in ("Shell Free", "Shell Sweep", "Shell Structured", "Solid Free", "Solid Sweep", "Solid Structured"):
            self.generationMethodMap[technique] = {}
            self.generationMethodListIdMap[technique] = {}

        self.materialList = "" # Set later on
        self.materialListId = None # Set later on
        self.geometricOrderListId = None # Set later on
        self.generationMethodMap["Shell Free"]["Quad Shell"] = ("Direct",)
        self.generationMethodMap["Shell Sweep"]["Quad Shell"] = ("Direct",)
        self.generationMethodMap["Shell Structured"]["Quad Shell"] = ("Direct",)

        self.generationMethodMap["Solid Free"]["Tet"] = ("Direct", "Virtual Topology", "Automatic")
        self.generationMethodMap["Solid Free"]["Hex"] = ("Virtual Topology", "Extrude")
        self.generationMethodMap["Solid Free"]["Quad Shell"] = ("Direct", )

        self.generationMethodMap["Solid Sweep"]["Hex"] = ("Direct", "Virtual Topology", "Extrude")
        self.generationMethodMap["Solid Sweep"]["Tet"] = ("Direct", "Virtual Topology")
        self.generationMethodMap["Solid Sweep"]["Quad Shell"] = ("Direct",)
        
        self.generationMethodMap["Solid Structured"]["Hex"] = ("Direct", "Virtual Topology", "Extrude")
        self.generationMethodMap["Solid Structured"]["Tet"] = ("Direct", "Virtual Topology")
        self.generationMethodMap["Solid Structured"]["Quad Shell"] = ("Direct",)

        # Formulations
        self.formulationMap = {}
        self.formulationListIdMap = {} # map = {technique:{elemShape:{generationMethod:{geometric_order:{(lisId1, listId2, ...)}}}}}
        for technique in ("Shell Free", "Shell Sweep", "Shell Structured", "Solid Free", "Solid Sweep", "Solid Structured"):
            self.formulationMap[technique] = {}
            self.formulationListIdMap[technique] = {}
            for elemShape in ("Quad Shell", "Tet", "Hex"):
                self.formulationMap[technique][elemShape] = {}
                self.formulationListIdMap[technique][elemShape] = {}
                if elemShape in self.generationMethodMap[technique].keys():
                    for method in self.generationMethodMap[technique][elemShape]:
                        self.formulationMap[technique][elemShape][method] = {}
                        self.formulationListIdMap[technique][elemShape][method] = {}
        self.formulationMap["Shell Free"]["Quad Shell"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration")
        self.formulationMap["Shell Free"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )
        self.formulationMap["Shell Sweep"]["Quad Shell"]["Direct"]["Linear"]  = ("Reduced Integration", "Full Integration")
        self.formulationMap["Shell Sweep"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )
        self.formulationMap["Shell Structured"]["Quad Shell"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration")
        self.formulationMap["Shell Structured"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )

        #
        self.formulationMap["Solid Free"]["Tet"]["Direct"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        self.formulationMap["Solid Free"]["Tet"]["Virtual Topology"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        self.formulationMap["Solid Free"]["Tet"]["Automatic"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Free"]["Tet"]["Direct"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                           "Full Integration, Hybrid", "Modified Integration, Hybrid")
        self.formulationMap["Solid Free"]["Tet"]["Virtual Topology"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                                     "Full Integration, Hybrid", "Modified Integration, Hybrid")
        self.formulationMap["Solid Free"]["Tet"]["Automatic"]["Quadratic"] =  ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                               "Full Integration, Hybrid", "Modified Integration, Hybrid")
        #
        self.formulationMap["Solid Free"]["Hex"]["Virtual Topology"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                                  "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        self.formulationMap["Solid Free"]["Hex"]["Extrude"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                         "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        #
        self.formulationMap["Solid Free"]["Hex"]["Virtual Topology"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        self.formulationMap["Solid Free"]["Hex"]["Extrude"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Free"]["Quad Shell"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration")
        # 
        self.formulationMap["Solid Free"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )
        #
        self.formulationMap["Solid Sweep"]["Hex"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                         "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        self.formulationMap["Solid Sweep"]["Hex"]["Virtual Topology"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                                   "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        self.formulationMap["Solid Sweep"]["Hex"]["Extrude"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                          "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        #
        self.formulationMap["Solid Sweep"]["Hex"]["Direct"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        self.formulationMap["Solid Sweep"]["Hex"]["Virtual Topology"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        self.formulationMap["Solid Sweep"]["Hex"]["Extrude"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Sweep"]["Tet"]["Direct"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        self.formulationMap["Solid Sweep"]["Tet"]["Virtual Topology"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Sweep"]["Tet"]["Direct"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                            "Full Integration, Hybrid", "Modified Integration, Hybrid")
        self.formulationMap["Solid Sweep"]["Tet"]["Virtual Topology"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                                      "Full Integration, Hybrid", "Modified Integration, Hybrid")
        #
        self.formulationMap["Solid Sweep"]["Quad Shell"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration")
        #
        self.formulationMap["Solid Sweep"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )
        #
        self.formulationMap["Solid Structured"]["Hex"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", 
                                                                              "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        self.formulationMap["Solid Structured"]["Hex"]["Virtual Topology"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                                        "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        self.formulationMap["Solid Structured"]["Hex"]["Extrude"]["Linear"] = ("Reduced Integration", "Full Integration", "Incompatible Modes", "Continuum Shell",
                                                                               "Reduced Integration, Hybrid", "Full Integration, Hybrid", "Incompatible Modes, Hybrid")
        #
        self.formulationMap["Solid Structured"]["Hex"]["Direct"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        self.formulationMap["Solid Structured"]["Hex"]["Virtual Topology"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        self.formulationMap["Solid Structured"]["Hex"]["Extrude"]["Quadratic"] = ("Reduced Integration", "Full Integration", "Reduced Integration, Hybrid", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Structured"]["Tet"]["Direct"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        self.formulationMap["Solid Structured"]["Tet"]["Virtual Topology"]["Linear"] = ("Full Integration", "Full Integration, Hybrid")
        #
        self.formulationMap["Solid Structured"]["Tet"]["Direct"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                                 "Full Integration, Hybrid", "Modified Integration, Hybrid")
        self.formulationMap["Solid Structured"]["Tet"]["Virtual Topology"]["Quadratic"] = ("Full Integration", "Modified Integration", "Improved Surface Integration",
                                                                                           "Full Integration, Hybrid", "Modified Integration, Hybrid")
        #
        self.formulationMap["Solid Structured"]["Quad Shell"]["Direct"]["Linear"] = ("Reduced Integration", "Full Integration")
        #
        self.formulationMap["Solid Structured"]["Quad Shell"]["Direct"]["Quadratic"] = ("Reduced Integration", )

        
        self.toolset = self.form.getOwner()

        #FXMAPFUNC(self, SEL_SELECTED, self.ID_TABLE, PmeModelDB.onCmdCheckSort)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_TABLE, PmeModelDB.onCmdTable)
        FXMAPFUNC(self, SEL_CLICKED, self.ID_TABLE, PmeModelDB.onCmdLeftClick)
        FXMAPFUNC(self, SEL_RIGHTBUTTONPRESS, self.ID_TABLE, PmeModelDB.onCmdRightClick)

        FXMAPFUNC(self, SEL_COMMAND, self.ID_MESH, PmeModelDB.onCmdMesh)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_SKIP, PmeModelDB.onCmdSkip)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_HEX, PmeModelDB.onCmdHex)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_QUAD_SHELL, PmeModelDB.onCmdQuadShell)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_CONTINUUM_SHELL, PmeModelDB.onCmdContinuumShell)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_TET, PmeModelDB.onCmdTet)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_DIRECT, PmeModelDB.onCmdDirect)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_VIRTUAL_TOPOLOGY, PmeModelDB.onCmdVirtualTopology)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_EXTRUDE, PmeModelDB.onCmdExtrude)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_AUTOMATIC, PmeModelDB.onCmdAutomatic)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_LINEAR, PmeModelDB.onCmdLinear)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_QUADRATIC, PmeModelDB.onCmdQuadratic)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_RIGID_ON, PmeModelDB.onCmdRigidOn)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_RIGID_OFF, PmeModelDB.onCmdRigidOff)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_SKIN_ON, PmeModelDB.onCmdSkinOn)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_SKIN_OFF, PmeModelDB.onCmdSkinOff)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_SKIN_PICK_SURFACE, PmeModelDB.onCmdSkinPickSurface)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_EDIT_VISIBLE, PmeModelDB.onCmdEditVisible)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_REDUCED_INTEGRATION , PmeModelDB.onCmdReducedIntegration)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_FULL_INTEGRATION , PmeModelDB.onCmdFullIntegration)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_INCOMPATIBLE , PmeModelDB.onCmdIncompatible)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_CONTINUUM_SHELL , PmeModelDB.onCmdContinuumShell)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_MODIFIED , PmeModelDB.onCmdModified)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_IMPROVED, PmeModelDB.onCmdImproved)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_REDUCED_INTEGRATION_HYBRID , PmeModelDB.onCmdReducedHybrid)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_FULL_INTEGRATION_HYBRID , PmeModelDB.onCmdFullIntegrationHybrid)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_INCOMPATIBLE_HYBRID , PmeModelDB.onCmdIncompatibleHybrid)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_MODIFIED_HYBRID , PmeModelDB.onCmdModifiedHybrid)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_RESUME , PmeModelDB.onCmdResume)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_SUPPRESS , PmeModelDB.onCmdSuppress)
        FXMAPFUNC(self, SEL_COMMAND, self.ID_PICK_SURFACE, PmeModelDB.onCmdSurface)
  
        kMat = 0
        materialNames = mdb.models[self.form.modelName].materials.keys()
        materialNames.sort()
        for materialName in materialNames:
            FXMAPFUNC(self, SEL_COMMAND, self.ID_LAST+kMat, PmeModelDB.onCmdMaterial)
            kMat = kMat + 1

        vf = FXHorizontalFrame(self, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK|FRAME_SUNKEN)
        numCols = len(defaultColIdToColTypeMap.keys())
        self.table = AFXTable(vf, 0, 0, 15, numCols, self, self.ID_TABLE, 
                              LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK|FRAME_SUNKEN|
                              AFXTABLE_NORMAL|AFXTABLE_EDITABLE|AFXTABLE_RESIZE|AFXTABLE_COLUMN_RESIZABLE)
        self.table.setLeadingRows(1)
        self.table.showHorizontalGrid(True)
        self.table.showVerticalGrid(True)
        labels = ""

        getAFXAliasMap().setName(self.table, 'pmeTable')

        for kcol, item in defaultColIdToColTypeMap.items():
            title, colType, colWidth, colJustify, isEditable = tableFeatures[item]
            self.table.setColumnType(kcol, colType)
            self.table.setColumnJustify(kcol, colJustify)
            self.table.setColumnWidth(kcol, colWidth)
            if isEditable:
                self.table.setColumnEditable(kcol, True)
            else:
                self.table.setColumnEditable(kcol, False)
            if not labels:
                labels = title
            else:
                labels = labels + "\t" + title
            self.table.setColumnSortable(kcol, True)
        self.table.setLeadingRowLabels(labels)
        self.table.setWidth(400)
        self.setTableLists()
        self.fillMaterialList()
        self.fillGeometricOrderList()
        self.table.setStretchableColumn(numCols-1)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setTableLists(self):
        def _getListString(items):
            listString = ""
            for k in range(len(items)):
                item = items[k]
                if k > 0: listString = listString + "\t"
                listString = listString + item
            return listString
        # Set Elem Shape lists
        for technique, elemShapes in self.elemShapeMap.items():
            self.elemShapeListIdMap[technique] =  self.table.addList(_getListString(elemShapes))
        # Set method lists
        for technique, elemMap in self.generationMethodMap.items():
            for elemShape, methods in elemMap.items():
                self.generationMethodListIdMap[technique][elemShape] = self.table.addList(_getListString(methods))
        # Set Formulation lists
        for technique, elemMap in self.formulationMap.items():
            for elemShape, methodMap in elemMap.items():
                for method, orderMap in methodMap.items():
                    for order, formulations in orderMap.items():
                        self.formulationListIdMap[technique][elemShape][method][order] = self.table.addList(_getListString(formulations))
        # Geometric order list
        orderList = "Linear\tQuadratic"
        self.geometricOrderListId = self.table.addList(orderList)
                
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def fillGeometricOrderList(self):
        # Set geoemtric order lists
        if self.currentColTypeToColIdMap.has_key(GEOMETRICORDER):
            kcol = self.currentColTypeToColIdMap[GEOMETRICORDER]
            self.table.setColumnType(kcol, AFXTable.LIST)
            self.table.setColumnListId(kcol, self.geometricOrderListId )
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def fillMaterialList(self):
        model = mdb.models[self.form.modelName]
        self.materials = {}
        materialList = ''
        if hasattr(model, 'materials'):
            if  len(model.materials.keys()) > 0:
                kMat = 0
                for matName in model.materials.keys():
                    self.materials[matName] = kMat
                    kMat = kMat + 1
                    if kMat > 1:
                        materialList = materialList + '\t'
                    materialList = materialList + "%s" % matName
            else:
                materialList = "<None Available>"
            if materialList != self.materialList:
                self.materialList = materialList
                self.materialListId = self.table.addList(self.materialList)
            if self.currentColTypeToColIdMap.has_key(MATERIAL):
                kcol = self.currentColTypeToColIdMap[MATERIAL]
                self.table.setColumnType(kcol, AFXTable.LIST)
                self.table.setColumnListId(kcol, self.materialListId)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getItemValue(self, krow, columnType):
        if columnType in self.currentColTypeToColIdMap.keys():
            # Column is visible, read from table
            kcol = self.currentColTypeToColIdMap[columnType]
            valuType = tableFeatures[columnType][1]
            if valuType == AFXTable.BOOL:
                return self.table.getItemBoolValue(krow,kcol)
            else:
                return self.table.getItemText(krow,kcol)
        else:
            # Column is hidden, read from stored info
            partName = self.rowToPartNameMap[krow]
            return self.hiddenTableMap[partName][columnType]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setItemValue(self, krow, columnType, value):
        if columnType in self.currentColTypeToColIdMap.keys():
            # Column is visible, write to table
            kcol = self.currentColTypeToColIdMap[columnType]
            valueType = tableFeatures[columnType][1]
            if valueType == AFXTable.BOOL:
                self.table.setItemBoolValue(krow,kcol,value)
            elif valueType in (AFXTable.FLOAT, AFXTable.INT):
                if columnType in (WARNINGS,ERRORS):
                    value = float(value)
                    if value < .1 and value > 0.0:
                        value = .1
                    self.table.setItemText(krow,kcol,'%*.*f' % (4,1,value))
                else:
                    self.table.setItemText(krow,kcol,'%s' % value)
            else:
                self.table.setItemText(krow,kcol,value)
        else:
            # Column is hidden, read from stored info
            partName = self.rowToPartNameMap[krow]
            self.hiddenTableMap[partName][columnType] = value

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRequiredSurfacePartNames(self):
        
        from kernelAccess import mdb
        partNames = []
        for krow in range(1, self.table.getNumRows()):
            partName = self.getItemValue(krow, NAME)
            technique = self.getItemValue(krow, TECHNIQUE)
            elemShape = self.getItemValue(krow, ELEMSHAPE)
            generationMethod = self.getItemValue(krow, GENERATIONMETHOD)
            if technique in ("Solid Free", "Solid Sweep", "Solid Structured"):
                if elemShape == "Quad Shell" or generationMethod == "Extrude":
                    p = mdb.models[self.form.modelName].parts[partName]
                    if not p.surfaces.has_key("PME_BotSurf"):
                        partNames.append(partName)
        return partNames
       
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSelections(self):
        numRows = self.table.getNumRows()
        selections = []
        if numRows > 1:
            for i in range(1, numRows):
                if self.table.isAnyItemInRowSelected(i):
                    selections.append(self.getItemValue(i, NAME))
        return selections
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getUniqueSelections(self, colType):
        selections = []
        for krow in range(1, self.table.getNumRows()):
            if self.table.isAnyItemInRowSelected(krow):
                selection =  self.getItemValue(krow, colType)
                if selection not in selections:
                    selections.append(selection)
        return selections

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSelectedTechniques(self):
        techniques = []
        for krow in range(1, self.table.getNumRows()):
            technique  = self.getItemValue(krow, TECHNIQUE)
            if self.table.isAnyItemInRowSelected(krow):
                if technique not in techniques:
                    techniques.append(technique)
        return techniques

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSelectedElemTypes(self):
        elemShapes = []
        for krow in range(1,self.table.getNumRows()):
            elemShape = self.getItemValue(krow, ELEMSHAPE)
            if self.table.isAnyItemInRowSelected(krow):
                if elemShape not in elemShapes:
                    elemShapes.append(elemShape)
        return elemShapes

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdMesh(self, sender, sel, ptr):
    
        from kernelAccess import mdb
        statusMap = {}
        for row in mdb.customData.pmeModels[self.form.modelName].table:
            statusMap[row[1]] = row[0]
        for krow in self.getItemsSelected()[STATUS]:
            status = self.getItemValue(krow, STATUS)
            partName = self.getItemValue(krow, NAME)
            if partName in statusMap.keys():
                origStatus = statusMap[partName]
            else:
                origStatus = ''
            if status == "Skip":
                self.setItemValue(krow, STATUS, origStatus)
        self.setStatusIcons()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSkip(self, sender, sel, ptr):
        for krow in self.getItemsSelected()[STATUS]:
            status = self.getItemValue(krow, STATUS)
            self.setItemValue(krow, STATUS, "Skip")
        self.setStatusIcons()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdHex(self, sender, sel, ptr):
        self.setItems(ELEMSHAPE, "Hex")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdQuadShell(self, sender, sel, ptr):
        self.setItems(ELEMSHAPE, "Quad Shell")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdContinuumShell(self, sender, sel, ptr):
        self.setItems(ELEMSHAPE, "Continuum Shell")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdTet(self, sender, sel, ptr):
        self.setItems(ELEMSHAPE, "Tet")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdDirect(self, sender, sel, ptr):
        self.setItems(GENERATIONMETHOD, "Direct")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdVirtualTopology(self, sender, sel, ptr):
        self.setItems(GENERATIONMETHOD, "Virtual Topology")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdExtrude(self, sender, sel, ptr):
        self.setItems(GENERATIONMETHOD, "Extrude")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdAutomatic(self, sender, sel, ptr):
        self.setItems(GENERATIONMETHOD, "Automatic")
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdLinear(self, sender, sel, ptr):
        self.setItems(GEOMETRICORDER, "Linear")
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdQuadratic(self, sender, sel, ptr):
        self.setItems(GEOMETRICORDER, "Quadratic")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdReducedIntegration(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Reduced Integration")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdFullIntegration(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Full Integration")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdIncompatible(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Incompatible Modes")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdContinuumShell(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Continuum Shell")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdModified(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Modified Integration")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdImproved(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Improved Surface Integration")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdReducedHybrid(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Reduced Integration, Hybrid")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdFullIntegrationHybrid(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Full Integration, Hybrid")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdIncompatibleHybrid(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Incompatible Modes, Hybrid")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdModifiedHybrid(self, sender, sel, ptr):
        self.setItems(FORMULATION, "Modified Integration, Hybrid")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdResume(self, sender, sel, ptr):
        partNames = self.getSelections()
        self.toolset.pmeModelGenericForm.setValues(self.form.modelName, partNames, "resumeInstances")
        self.toolset.pmeModelGenericForm.issueCommands()
        self.updateStatuses()
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSuppress(self, sender, sel, ptr):
        partNames = self.getSelections()
        self.toolset.pmeModelGenericForm.setValues(self.form.modelName, partNames, "suppressInstances")
        self.toolset.pmeModelGenericForm.issueCommands()
        self.updateStatuses()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdRigidOn(self, sender, sel, ptr):
        self.setItems(RIGID, True)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdRigidOff(self, sender, sel, ptr):
        self.setItems(RIGID, False)
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSkinOn(self, sender, sel, ptr):
        rows = self.getItemsSelected()[SKIN]
        for krow in rows:
            technique = self.getItemValue(krow, TECHNIQUE)
            if technique not in ("Shell Structured", "Shell Sweep", "Shell Free"):
                if not self.getItemValue(krow, RIGID):
                    self.setItemValue(krow, SKIN, True)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSkinOff(self, sender, sel, ptr):
        self.setItems(SKIN, False)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSkinPickSurface(self, sender, sel, ptr):
        validPartNames = []
        for partName in self.getSelections():
            if self.getItemValue(self.partNameToRowMap[partName], ELEMSHAPE) != "Quad Shell":
                validPartNames.append(partName)
        if len(validPartNames) > 0:
            modelName = self.form.modelName
            cmd = "p = mdb.models['%s'].parts['%s']\n" % (modelName, validPartNames[0])
            cmd = cmd + "session.viewports[session.currentViewportName].setValues(displayedObject=p)\n"
            sendCommand(cmd)
            self.toolset.pmeSurfaceProcedure.setValues(self, self.form.modelName, validPartNames, "createSkinSurf", "PME_OuterSkin")
            self.toolset.pmeSurfaceProcedure.activate()
        else:
            msg = "Skins can not be applied to 'Quad Shell' shaped elements ."
            showAFXErrorDialog(getAFXApp().getAFXMainWindow(), msg)                
            return 0
        return 1

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdSurface(self, sender, sel, ptr):
        
        requiredPartNames = self.getRequiredSurfacePartNames()
        for partName in self.getSelections():
            if partName not in requiredPartNames:
                requiredPartNames.append(partName)
        modelName = self.form.modelName
        cmd = "p = mdb.models['%s'].parts['%s']\n" % (modelName, requiredPartNames[0])
        cmd = cmd + "session.viewports[session.currentViewportName].setValues(displayedObject=p)\n"
        sendCommand(cmd)
        self.toolset.pmeSurfaceProcedure.setValues(self, self.form.modelName, requiredPartNames, "createBotSurf", "PME_BotSurf")
        self.toolset.pmeSurfaceProcedure.activate()
        return 1

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getFullItemsSelected(self):
        rows = []; cols = []
        for krow in range(1,self.table.getNumRows()):
            if self.table.isRowSelected(krow):
                rows.append(krow)
        for kcol in range(len(self.currentColIdToColTypeMap)):
             if self.table.isColumnSelected(kcol):
                cols.append(kcol)
        return rows,cols

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getItemsSelected(self):
        colMap = {}
        for kcol in range(len(self.currentColIdToColTypeMap)):
            if self.table.isAnyItemInColumnSelected(kcol):
                colType = self.currentColIdToColTypeMap[kcol]
                if colType not in colMap.keys(): colMap[colType] = []
                for krow in range(1,self.table.getNumRows()):
                    if self.table.isItemSelected(krow,kcol):
                        colMap[colType].append(krow)
        return colMap

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setItems(self, colType, value):
        for krow in self.getItemsSelected()[colType]:
            self.setItemValue(krow, colType, value)
        self.onCmdTable()
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdMaterial(self, sender, sel, ptr):
        materialNames = mdb.models[self.form.modelName].materials.keys()
        materialNames.sort()
        kIndex = SELID(sel) - self.ID_LAST
        materialName = materialNames[kIndex]
        for krow in self.getItemsSelected()[MATERIAL]:
            self.setItemValue(krow, MATERIAL, materialName)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdLeftClick(self, sender, sel, ptr):
        "This function is called everytime a user clicks on the table"

        # Find mouse y position and get row.  If top row, sort
        #
        self.sortColumn = self.table.getColumnAtX(self.table.getCursorPosition()[1])
        row = self.table.getRowAtY(self.table.getCursorPosition()[2])
        if row == 0:
            self.sortOrder = self.table.getColumnSortOrder(self.sortColumn)
            self.sortTable()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdRightClick(self, sender, sel, ptr):

        # menu = [ ((text, enabled_flag, sel), ...), (....)]
        #   each group separated by AFXSeparator

        self.table.setPopupOptions(opts=AFXTable.POPUP_NONE)
        self.table.clearClientPopupItems()
        fullRows, fullColumns = self.getFullItemsSelected()
        menu = ()
        showPopUpMenu = 1
        if len(fullRows) > 0:
            menu = ( (("Skip", True, self.ID_SKIP),
                     ("Don't Skip", True, self.ID_MESH), ),
                     )
        else:
            colMap = self.getItemsSelected()
            
            # If there are items in table
            #
            if len(colMap.keys()) != 0:
            
                # More than 1 items selected
                #
                if len(colMap.keys()) > 1: return


                colType = colMap.keys()[0]
                
                # Variable to show pop up menu. Its available in all 
                # menus except for the ones that pop up a dialog: Element size and thickness
                #
                if colType == STATUS:
                    menu = ( (("Skip", True, self.ID_SKIP),
                              ("Don't Skip", True, self.ID_MESH), ),
                              )
                    
                    # Supress/resume for STATUS, NAME and TECHNIQUE only
                    menu += (
                                (("Suppress Instances", True, self.ID_SUPPRESS),("Resume Instances", True, self.ID_RESUME), ),
                            )

                elif colType in (NAME,TECHNIQUE):
                    menu = ()
                    menu += (
                                (("Suppress Instances", True, self.ID_SUPPRESS),("Resume Instances", True, self.ID_RESUME), ),
                            )
                elif colType == ELEMSHAPE:
                    canHex=True; canQuadShell=True; canContShell=True; canTet=True
                    for technique in self.getSelectedTechniques():
                        if "Hex" not in self.elemShapeMap[technique]: canHex=False
                        if "Quad Shell" not in self.elemShapeMap[technique]: canQuadShell=False
                        if "Tet" not in self.elemShapeMap[technique]: canTet=False
                    menu = ( (("Hex", canHex, self.ID_HEX),
                              ("Quad Shell", canQuadShell, self.ID_QUAD_SHELL),
                              ("Tet", canTet, self.ID_TET) ), ) 
                elif colType == FORMULATION:
                    canReduced=True; canFull=True; canIncompatible=True; canContShell=True; canModified=True; canImproved=True
                    canReducedHybrid=True; canFullHybrid=True; canIncompatibleHybrid=True; canModifiedHybrid=True
                    for krow in range(1, self.table.getNumRows()):
                        if self.table.isAnyItemInRowSelected(krow):
                            technique = self.getItemValue(krow, TECHNIQUE)
                            elemshape = self.getItemValue(krow, ELEMSHAPE)
                            order = self.getItemValue(krow, GEOMETRICORDER)
                            method = self.getItemValue(krow, GENERATIONMETHOD)
                            if "Reduced Integration" not in self.formulationMap[technique][elemshape][method][order]: canReduced=False
                            if "Full Integration" not in self.formulationMap[technique][elemshape][method][order]: canFull=False
                            if "Incompatible Modes" not in self.formulationMap[technique][elemshape][method][order]: canIncompatible=False
                            if "Continuum Shell" not in self.formulationMap[technique][elemshape][method][order]: canContShell=False
                            if "Modified Integration" not in self.formulationMap[technique][elemshape][method][order]: canModified=False
                            if "Improved Surface Integration" not in self.formulationMap[technique][elemshape][method][order]: canImproved=False
                            if "Reduced Integration, Hybrid" not in self.formulationMap[technique][elemshape][method][order]: canReducedHybrid=False
                            if "Full Integration, Hybrid" not in self.formulationMap[technique][elemshape][method][order]: canFullHybrid=False
                            if "Incompatible Modes, Hybrid" not in self.formulationMap[technique][elemshape][method][order]: canIncompatibleHybrid=False
                            if "Modified Integration, Hybrid" not in self.formulationMap[technique][elemshape][method][order]: canModifiedHybrid=False
                    menu = ( (("Reduced Integration", canReduced, self.ID_REDUCED_INTEGRATION),
                              ("Full Integration", canFull, self.ID_FULL_INTEGRATION),
                              ("Incompatible Modes", canIncompatible, self.ID_INCOMPATIBLE),
                              ("Continuum Shell", canContShell, self.ID_CONTINUUM_SHELL),
                              ("Modified Integration", canModified, self.ID_MODIFIED),
                              ("Improved Surface Integration", canImproved, self.ID_IMPROVED)),
                             (("Reduced Integration, Hybrid", canReducedHybrid, self.ID_REDUCED_INTEGRATION_HYBRID),
                              ("Full Integration, Hybrid", canFullHybrid, self.ID_FULL_INTEGRATION_HYBRID),
                              ("Incompatible Modes, Hybrid", canIncompatibleHybrid, self.ID_INCOMPATIBLE_HYBRID),
                              ("Modified Integration, Hybrid", canModifiedHybrid, self.ID_MODIFIED_HYBRID)
                              ) )                           
                elif colType == GENERATIONMETHOD:
                    canDirect=True; canVirtTop=True; canExtrude=True; canAutomatic=True;
                    for krow in range(1,self.table.getNumRows()):
                        if self.table.isAnyItemInRowSelected(krow):
                            technique = self.getItemValue(krow, TECHNIQUE)
                            elemtype = self.getItemValue(krow, ELEMSHAPE)
                            methods = self.generationMethodMap[technique][elemtype]
                            if "Direct" not in methods: canDirect=False
                            if "Virtual Topology" not in methods: canVirtTop=False
                            if "Extrude" not in methods:  canExtrude=False
                            if "Automatic" not in methods: canAutomatic=False
                    menuList = []
                    if canDirect: menuList.append( ("Direct", True, self.ID_DIRECT) )
                    if canVirtTop: menuList.append( ("Virtual Topology", True, self.ID_VIRTUAL_TOPOLOGY) )
                    if canExtrude: menuList.append( ("Extrude", True, self.ID_EXTRUDE) )
                    if canAutomatic: menuList.append( ("Automatic", True, self.ID_AUTOMATIC) )
                    menu = ( tuple(menuList), )
                elif colType == GEOMETRICORDER: 
                     menu = ( (("Linear", True, self.ID_LINEAR),
                               ("Quadratic", True, self.ID_QUADRATIC)),
                               )
                elif colType == ELEMSIZE:
                    partNames = self.getSelections()
                    if not self.textDB:
                        self.textDB = PmeTextFieldDB()
                        self.textDB.create()
                    sizeCol = self.currentColTypeToColIdMap[ELEMSIZE]
                    self.textDB.initialize(self, sizeCol, partNames, 'Element Size', 'Element Size:')
                    self.textDB.show()
                    status, x, y, buttons = self.textDB.getRoot().getCursorPosition()
                    self.textDB.position(x, y, 200, 75)
                    showPopUpMenu = 0
                elif colType == MATERIAL:
                    partNames = self.getSelections()
                    materialNames = mdb.models[self.form.modelName].materials.keys()
                    materialNames.sort()
                    if len(materialNames) > 20:
                        matCol = self.currentColTypeToColIdMap[MATERIAL]
                        db = PmeComboBoxDB()
                        db.create()
                        db.initialize(self, matCol, partNames, 'Material Assignments', 'Material Name:', materialNames)
                        db.show()
                    else:
                        menuList = []
                        kMat = 0
                        for materialName in materialNames:
                            menuList.append(  (materialName, True, self.ID_LAST+kMat)  )
                            kMat = kMat + 1
                        menu = (tuple(menuList), )
                elif colType == RIGID:
                    menu = ( (("On", True, self.ID_RIGID_ON),
                              ("Off", True, self.ID_RIGID_OFF),
                              ),   )  
                elif colType == SKIN:
                    menu = ( (("On", True, self.ID_SKIN_ON),
                              ("Off", True, self.ID_SKIN_OFF),
                              ("Pick Skin Surface", True, self.ID_SKIN_PICK_SURFACE),
                              ),   )
                elif colType == THICKNESS:
                    partNames = self.getSelections()
                    if not self.textDB:
                        self.textDB = PmeTextFieldDB()
                        self.textDB.create()
                    thickCol = self.currentColTypeToColIdMap[THICKNESS]
                    self.textDB.initialize(self, thickCol, partNames, 'Shell Thickness', 'Shell Thickness:')
                    self.textDB.show()
                    status, x, y, buttons = self.textDB.getRoot().getCursorPosition()
                    self.textDB.position(x, y, 200, 75)
                    self.textDB.setFocus()
                    showPopUpMenu = 0
                elif colType in (THICKNESS,VOLUME,ERRORS,WARNINGS,MASS,TIMEINCREMENT):
                    menu = ()

        # Build menus
        #        
        canPickSurface = False
        for krow in range(1,self.table.getNumRows()):
            if self.table.isAnyItemInRowSelected(krow):
                technique = self.getItemValue(krow, TECHNIQUE)
                if technique in ("Shell Sweep", "Shell Free", "Shell Structured"): continue
                elemShape = self.getItemValue(krow, ELEMSHAPE)
                method = self.getItemValue(krow, GENERATIONMETHOD)
                if elemShape == "Quad Shell" or method == "Extrude":
                    canPickSurface = True
                    break
        if canPickSurface:
            menu += (
                        (("Pick Bottom Surface", canPickSurface, self.ID_PICK_SURFACE), ),
                    )

        if showPopUpMenu:
        
            # Create menu
            #
            fullMenu = menu + ((("Edit Visible Columns", True, self.ID_EDIT_VISIBLE), ), )
            mainWindow = getAFXApp().getAFXMainWindow()
            menuPane = FXMenuPane(self)
            PmeModelDB.menuPaneItr += 1
            
            kItem = 0
            for kGroup in range(len(fullMenu)):
                group = fullMenu[kGroup]
                if kGroup > 0:
                    FXHorizontalSeparator(menuPane)
                for item in group:
                    kItem = kItem + 1
                    title, enable, itemSel = item
                    mc = FXMenuCommand(menuPane, title, None, self, itemSel)                
                    if not enable:
                        mc.disable()
                    itemName = 'Item_%d' % kItem
                    getAFXAliasMap().setName(mc, itemName)
                    setattr(menuPane, itemName, mc)
            menuPane.create()
            status, x, y, buttons = mainWindow.getCursorPosition()
            x, y = mainWindow.translateCoordinatesTo( mainWindow.getRoot(), x, y )
            menuPane.popup(None, x, y)
            getAFXAliasMap().setPrefix(menuPane, 'PME RMB menu-%s' % PmeModelDB.menuPaneItr)
 
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdEditVisible(self, sender=-1, sel=-1, ptr=-1):
        visibleDB = PmeVisibleColumnsDB(self, fullColIdToColTypeMap, tableFeatures, self.currentColTypeToColIdMap)
        visibleDB.create()
        visibleDB.showModal()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getCurrentFullTable(self):
        tmpKw = []
        for krow in range(1,self.table.getNumRows()):
            tmpRow = []
            for kcol in range(len(fullColIdToColTypeMap.keys())):
                colType = fullColIdToColTypeMap[kcol]
                tmpRow.append(self.getItemValue(krow,colType))
            tmpKw.append(tmpRow)
        return tmpKw
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def changeColumnArrangement(self, currentColIdToColTypeMap):

        # Store a backup
        backup = self.getCurrentFullTable()
        sortColumn = self.table.getCurrentSortColumn()
        if sortColumn in self.currentColTypeToColIdMap.keys():
            sortType = self.currentColIdToColTypeMap[sortColumn]
        else:
            sortType = None
        # Set new column maps
        prevColTypeToColIdMap = copy.deepcopy(self.currentColTypeToColIdMap)
        self.currentColIdToColTypeMap = currentColIdToColTypeMap
        self.currentColTypeToColIdMap = getReverse(self.currentColIdToColTypeMap)
        # Resize table, change column settings, and update titles
        numRows = self.table.getNumRows()
        numCols = len(self.currentColIdToColTypeMap.keys())
        self.table.setTableSize(numRows,numCols)
        self.table.setStretchableColumn(numCols-1)
        labels = ""
        for kcol, item in self.currentColIdToColTypeMap.items():
            title, colType, colWidth, colJustify, isEditable = tableFeatures[item]
            self.table.setColumnType(kcol, colType)
            self.table.setColumnJustify(kcol, colJustify)
            self.table.setColumnWidth(kcol, colWidth)
            if isEditable:
                self.table.setColumnEditable(kcol, True)
            else:
                self.table.setColumnEditable(kcol, False)
            if not labels:
                labels = title
            else:
                labels = labels + "\t" + title
            self.table.setColumnSortable(kcol, True)
            for krow in range(1, self.table.getNumRows()):
                self.table.setItemBackColor(krow,kcol,'None')

        self.table.setLeadingRowLabels(labels)
        if sortType and sortType in self.currentColTypeToColIdMap.keys():
            self.table.setCurrentSortColumn(self.currentColTypeToColIdMap[sortType])
        self.fillMaterialList()
        self.fillGeometricOrderList()
        # Update values of table
        self.hiddenTableMap = {}
        for krow in range(1,self.table.getNumRows()):
            partName = self.rowToPartNameMap[krow]
            self.hiddenTableMap[partName] = {}
            for kcol in range(len(fullColIdToColTypeMap.keys())):
                colType = fullColIdToColTypeMap[kcol]                
                if colType == ELEMSHAPE:
                    technique = backup[krow-1][fullColTypeToColIdMap[TECHNIQUE]]
                    elemShape = backup[krow-1][fullColTypeToColIdMap[ELEMSHAPE]]
                    method = backup[krow-1][fullColTypeToColIdMap[GENERATIONMETHOD]]
                    order = backup[krow-1][fullColTypeToColIdMap[GEOMETRICORDER]]
                    formulation = backup[krow-1][fullColTypeToColIdMap[FORMULATION]]
                    self.setElemControls(krow, technique, elemShape, method, order, formulation)
                elif colType in (ELEMSHAPE, GENERATIONMETHOD, GEOMETRICORDER, FORMULATION):
                    pass # already filled above
                else:
                    self.setItemValue(krow, colType, backup[krow-1][kcol])
        self.updateTableState()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCmdTable(self, sender=-1, sel=-1, ptr=-1):
        self.checkSurfaces()
        self.updateTableState()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def checkSurfaces(self):
        names = []
        model = mdb.models[self.form.modelName]
        for krow in range(1, self.table.getNumRows()):
            technique = self.getItemValue(krow, TECHNIQUE)
            if technique in ("Shell Sweep", "Shell Free", "Shell Structured"): continue
            elemShape = self.getItemValue(krow, ELEMSHAPE)
            method = self.getItemValue(krow, GENERATIONMETHOD)
            if elemShape == "Quad Shell" or method == "Extrude":
                partName = self.getItemValue(krow, NAME)
                if not model.parts[partName].surfaces.has_key("PME_BotSurf"):
                    names.append(partName)
        if len(names) > 0:
            cmd = "p = mdb.models['%s'].parts['%s']\n" % (self.form.modelName, names[0])
            cmd = cmd + "session.viewports[session.currentViewportName].setValues(displayedObject=p)\n"
            sendCommand(cmd)
            self.toolset.pmeSurfaceProcedure.setValues(self, self.form.modelName, names, "createBotSurf", "PME_BotSurf")
            self.toolset.pmeSurfaceProcedure.activate()
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def resetTableLists(self):
        for krow in range(1,self.table.getNumRows()):
            technique = self.getItemValue(krow, TECHNIQUE)
            elemShape = self.getItemValue(krow, ELEMSHAPE)
            method = self.getItemValue(krow, GENERATIONMETHOD)
            order =  self.getItemValue(krow, GEOMETRICORDER)
            formulation = self.getItemValue(krow, FORMULATION)
            self.setElemControls(krow, technique, elemShape, method, order, formulation)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setElemControls(self, krow, technique, elemShape, method, order, formulation):
        hasElemShapeCol = ELEMSHAPE in self.currentColTypeToColIdMap.keys()
        hasMethodCol = GENERATIONMETHOD in self.currentColTypeToColIdMap.keys()
        hasFormCol = FORMULATION in self.currentColTypeToColIdMap.keys()
        availableElemTypes = self.elemShapeMap[technique]
        # Element Shape
        if elemShape not in availableElemTypes:
            elemShape = availableElemTypes[0]
        if hasElemShapeCol:
            elemShapeCol = self.currentColTypeToColIdMap[ELEMSHAPE]
            self.table.setItemListId(krow, elemShapeCol, self.elemShapeListIdMap[technique])
            if len(availableElemTypes) == 1:
                self.table.setItemEditable(krow,elemShapeCol,False)
            else:
                self.table.setItemEditable(krow,elemShapeCol,True)            
        self.setItemValue(krow, ELEMSHAPE, elemShape)
        # Generation Method
        availableMethods = self.generationMethodMap[technique][elemShape]
        if method not in availableMethods:
            method = availableMethods[0]
        if hasMethodCol:
            methodCol = self.currentColTypeToColIdMap[GENERATIONMETHOD]
            self.table.setItemListId(krow, methodCol, self.generationMethodListIdMap[technique][elemShape])
            if len(availableMethods) == 1:
                self.table.setItemEditable(krow,methodCol,False)
            else:
                self.table.setItemEditable(krow,methodCol,True)
        self.setItemValue(krow, GENERATIONMETHOD, method)
        # Geometric Order
        self.setItemValue(krow, GEOMETRICORDER, order)
        # Formulation
        availableFormulations = self.formulationMap[technique][elemShape][method][order]
        if formulation not in availableFormulations:
            formulation = availableFormulations[0]
        if hasFormCol:
            formCol = self.currentColTypeToColIdMap[FORMULATION]
            self.table.setItemListId(krow, formCol, self.formulationListIdMap[technique][elemShape][method][order])
            if len(availableFormulations) == 1:
                self.table.setItemEditable(krow,formCol,False)
            else:                
                self.table.setItemEditable(krow,formCol,True)
        self.setItemValue(krow, FORMULATION, formulation)
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSortedRows(self, rows):
        # rows = row ids (subset of all rows)
        # return list of row numbers based on alphabetical order of part name
        names = {}
        for krow in rows:
            names[self.getItemValue(krow, NAME)] = krow
        newRows = []
        sortedNames = names.keys()
        sortedNames.sort()
        for name in sortedNames:
            newRows.append(names[name])
        return newRows

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def sortTable(self):

        # Determine sortMap # {newRowNum:oldRowNum}
        sortColType = self.currentColIdToColTypeMap[self.sortColumn]
        values = {}
        for krow in range(1, self.table.getNumRows()):
            value = self.getItemValue(krow, sortColType)
            if tableFeatures[sortColType][1] == AFXTable.FLOAT:
                try:
                    value = float(value)
                except:
                    value=None
            elif tableFeatures[sortColType][1] == AFXTable.INT:
                try:
                    value = int(value)
                except:
                    value = None
            elif sortColType == STATUS:
                partName = self.getItemValue(krow, NAME)
                if not self.isPartActive(partName):
                    value = "Deactivated"                    
            if not values.has_key(value):
                values[value] = []
            values[value].append(krow)
        sortMap = {} # {newRowNum:oldRowNum}
        newRow = 1
        keys = values.keys()
        keys.sort()
        if self.sortOrder == 2: keys.reverse()
        for key in keys:
            rows = values[key]
            sortedRows = self.getSortedRows(rows)
            for row in sortedRows:
                sortMap[newRow] = row
                newRow = newRow + 1
        # Store current order
        tmpKw = self.getCurrentFullTable()
        # Update row_number/part_name maps
        newPartNameToRowMap = {}
        revSortMap = getReverse(sortMap)
        for krow in range(1, self.table.getNumRows()):
            partName = self.getItemValue(krow, NAME)
            newPartNameToRowMap[partName] = revSortMap[krow]
        self.partNameToRowMap = newPartNameToRowMap
        self.rowToPartNameMap = getReverse(newPartNameToRowMap)
        # Write new order
        for krow in range(1, self.table.getNumRows()):
            row = tmpKw[sortMap[krow]-1]
            for kcol,colType in fullColIdToColTypeMap.items():
                oldValue = row[kcol]
                if colType == ELEMSHAPE:
                    self.setElemControls(krow, row[fullColTypeToColIdMap[TECHNIQUE]], row[fullColTypeToColIdMap[ELEMSHAPE]], row[fullColTypeToColIdMap[GENERATIONMETHOD]],
                                         row[fullColTypeToColIdMap[GEOMETRICORDER]], row[fullColTypeToColIdMap[FORMULATION]])
                elif colType in (ELEMSHAPE, GENERATIONMETHOD, GEOMETRICORDER, FORMULATION):
                    pass # Done in setElementControls
                else:
                    self.setItemValue(krow, colType, oldValue)        
        self.updateTableState()
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateStatuses(self):
    
        from kernelAccess import mdb
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        for krow in range(0,len(pmeModel.table)):
            row = pmeModel.table[krow]
            status = row[0]
            partName = row[1]
            rowId = self.partNameToRowMap[partName]
            tableStatus = self.getItemValue(rowId, STATUS)
            if tableStatus != "Skip":
                self.setItemValue(rowId, STATUS, status)
        self.setStatusIcons()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateTableState(self):
        # The following routines are based entirely on the values within the table
        self.setTechniqueIcons()
        self.resetTableLists()
        self.setTableColors()
        self.setStatusIcons()
        self.setErrorsWarnings()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getUpdatedPartNames(self, fromTable):
    
        from kernelAccess import mdb
        model = mdb.models[self.form.modelName]
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        removedPartNames=[]; newPartNames=[]; existingPartNames=[]
        
        # existingPartNames - part names which still exist in model
        # Remove the ones with independent instances
        #
        mdbPartNames = []
        for key in model.parts.keys():
            if self.isInstanceDependent(model, key):
                mdbPartNames.append(key)
        if fromTable:
            for krow in range(1,self.table.getNumRows()):
                partName = self.getItemValue(krow, NAME)
                if partName in mdbPartNames:
                    existingPartNames.append(partName)
                else:
                    removedPartNames.append(partName)
        else:
            for row in pmeModel.table:
                partName = row[1]
                if partName in mdbPartNames:
                    existingPartNames.append(row[1])
                else:
                    removedPartNames.append(partName)

        # newPartNames - part names that have been added to model, but not put into mdb.customData.pmeModels[''].table
        newPartNames = []
        for partName in model.parts.keys():
            if partName not in existingPartNames:
                p = model.parts[partName]
                if not self.isInstanceDependent(model, partName):
                    continue # Skip parts with independent instances
                if len(p.cells) > 0 or len(p.faces) > 0: 
                    if partName[-3:] != partSuffix:
                        newPartNames.append(partName)
        return removedPartNames, newPartNames, existingPartNames

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def fillTableFromCustomData(self):
        
        from kernelAccess import mdb
        # Only called when dialog is first brought up
        self.rowToPartNameMap = {}
        self.partNameToRowMap = {}
        self.hiddenTableMap = {}
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        removedPartNames, newPartNames, existingPartNames = self.getUpdatedPartNames(fromTable=False)
        numParts = len(newPartNames) + len(existingPartNames)
        self.table.setTableSize(numParts+1,len(defaultColIdToColTypeMap))
        krow = 1
        # Generate defaults and store on session.customData, then write to top of table
        if len(newPartNames) > 0:
            cmd = "mdb.customData.pmeModels['%s'].updateNewPartInfo()\n" % self.form.modelName
            sendCommand(cmd)
            
            # Add new rows to end of table
            for row in session.customData.newPartInfo:
                self.updateTableRowFromCustomData(krow, row)
                krow = krow + 1
        # Fill lines from pmeModel
        for curRow in range(0,len(pmeModel.table)):
            row = pmeModel.table[curRow]
            if row[1] in removedPartNames: continue # part has been deleted
            self.updateTableRowFromCustomData(krow, row)
            krow = krow + 1
        self.sortTable()
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateTableRowFromCustomData(self, krow, row):
        
        # Row is completely filled and in default order
        from kernelAccess import mdb
        model = mdb.models[self.form.modelName]
        a = model.rootAssembly
        partName = row[1]        

        self.rowToPartNameMap[krow] = partName
        self.partNameToRowMap[partName] = krow
        self.hiddenTableMap[partName] = {}
        mdbPart = model.parts[partName]
        for kcol in range(len(fullColIdToColTypeMap.keys())):
            colType = fullColIdToColTypeMap[kcol]
            colValue = row[kcol]
            if colType == STATUS:
                if STATUS in self.currentColTypeToColIdMap.keys():
                    colId = self.currentColTypeToColIdMap[STATUS]
                    if self.table.getItemText(krow, colId) == "Skip":
                        colValue = "Skip"
            if colType == VOLUME:
                self.newVolumeMap[partName] = float(colValue)
            if colType == ELEMSHAPE:
                self.setElemControls(krow, row[fullColTypeToColIdMap[TECHNIQUE]], row[fullColTypeToColIdMap[ELEMSHAPE]], row[fullColTypeToColIdMap[GENERATIONMETHOD]],
                                     row[fullColTypeToColIdMap[GEOMETRICORDER]], row[fullColTypeToColIdMap[FORMULATION]])
            elif colType in (ELEMSHAPE, GENERATIONMETHOD, GEOMETRICORDER, FORMULATION):
                pass # Done in setElemControls
            else:
                self.setItemValue(krow, colType, colValue)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setTechniqueIcons(self):
        if TECHNIQUE not in self.currentColTypeToColIdMap.keys(): return
        kcol = self.currentColTypeToColIdMap[TECHNIQUE]
        for krow in range(1,self.table.getNumRows()):
            technique = self.getItemValue(krow, TECHNIQUE)
            if technique == "Shell Structured":
                self.table.setItemIcon(krow, kcol, shellStructuredIcon)
                self.table.setItemText(krow, kcol, "Shell Structured")
            elif technique == "Shell Sweep":
                self.table.setItemIcon(krow, kcol, shellSweepIcon)
                self.table.setItemText(krow, kcol, "Shell Sweep")
            elif technique == "Shell Free":
                self.table.setItemIcon(krow, kcol, shellFreeIcon)
                self.table.setItemText(krow, kcol, "Shell Free")
            elif technique == "Solid Structured":
                self.table.setItemIcon(krow, kcol, solidStructuredIcon)
                self.table.setItemText(krow, kcol, "Solid Structured")
            elif technique == "Solid Sweep":
                self.table.setItemIcon(krow, kcol,  solidSweptIcon)
                self.table.setItemText(krow, kcol, "Solid Sweep")
            else: # FREE
                self.table.setItemIcon(krow, kcol, solidFreeIcon)
                self.table.setItemText(krow, kcol, "Solid Free")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setTableColors(self):
    
        from kernelAccess import mdb
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        # Check if color is required
        hasSkinCol = SKIN in self.currentColTypeToColIdMap.keys()
        hasThickCol = THICKNESS in self.currentColTypeToColIdMap.keys()
        hasOrderCol = GEOMETRICORDER in self.currentColTypeToColIdMap.keys()
        # Loop over each part
        if hasSkinCol:
            skinCol = self.currentColTypeToColIdMap[SKIN]
        if hasThickCol:
            thickCol = self.currentColTypeToColIdMap[THICKNESS]
        if hasOrderCol:
            orderCol = self.currentColTypeToColIdMap[GEOMETRICORDER]
        
        for krow in range(1, self.table.getNumRows()):
            partName = self.getItemValue(krow, NAME)
            technique = self.getItemValue(krow, TECHNIQUE)
            self.getItemValue(krow, GENERATIONMETHOD)
            elemShape = self.getItemValue(krow, ELEMSHAPE)
            rigid = self.getItemValue(krow, RIGID)
                    
            # Checks for skin
            if rigid == True or elemShape == "Quad Shell":
                self.setItemValue(krow, SKIN, False)
                if hasSkinCol:
                    self.table.setItemEditable(krow,skinCol,False)
                    self.table.setItemBackColor(krow,skinCol,'Grey80')
            else:
                if hasSkinCol:
                    self.table.setItemEditable(krow,skinCol,True)
                    self.table.setItemBackColor(krow,skinCol,'None')
            # Checks for thickness
            if technique in ("Shell Free", "Shell Structured", "Shell Sweep"):
                if hasThickCol:
                    self.table.setItemEditable(krow,thickCol,True)                
                    self.table.setItemBackColor(krow,thickCol,'None')
            else:
                self.setItemValue(krow, THICKNESS, "")
                if hasThickCol:
                    self.table.setItemEditable(krow,thickCol,False)
                    self.table.setItemBackColor(krow,thickCol,'Grey80')
            
            # Order must be linear for continuum shells
            if elemShape == "Continuum Shell":
                self.setItemValue(krow, GEOMETRICORDER, "Linear")
                if hasOrderCol:
                    self.table.setItemEditable(krow,orderCol,False)
                else:
                    self.table.setItemEditable(krow,orderCol,True)
            # Calculate volume for shells
            if technique in ("Shell Free", "Shell Structured", "Shell Sweep"):
                try:
                    thickness = float(self.getItemValue(krow, THICKNESS))
                except:
                    continue
                if partName in pmeModel.shellThicknessesFromSolids.keys():
                    volume = thickness*pmeModel.shellThicknessesFromSolids[partName]
                    self.setItemValue(krow, VOLUME, volume)
                elif partName in self.newVolumeMap.keys():
                    volume = thickness*self.newVolumeMap[partName]
                    self.setItemValue(krow, VOLUME, volume)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def isPartActive(self, partName):
    
        from kernelAccess import mdb
        return mdb.customData.pmeModels[self.form.modelName].activationMap[partName]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setStatusIcons(self):
    
        from kernelAccess import mdb
    
        if STATUS not in self.currentColTypeToColIdMap.keys(): return
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        statusCol = self.currentColTypeToColIdMap[STATUS]
        for krow in range(1,self.table.getNumRows()):
            status = self.getItemValue(krow, STATUS)
            partName =self.getItemValue(krow, NAME)
            if status == "Skip":
                self.table.setItemIcon(krow,statusCol,None)
            elif not self.isPartActive(partName):
                self.table.setItemIcon(krow, statusCol, deactivatedIcon)
            elif status == "Y":
                self.table.setItemIcon(krow, statusCol, meshIcon)
            elif status == "N":
                self.table.setItemIcon(krow,statusCol, crossIcon)
            else:
                self.table.setItemIcon(krow,statusCol,None)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setErrorsWarnings(self):
        hasErrorsCol = ERRORS in self.currentColTypeToColIdMap.keys()
        hasWarningsCol = WARNINGS in self.currentColTypeToColIdMap.keys()
        
        from kernelAccess import mdb
        
        pmeModel = mdb.customData.pmeModels[self.form.modelName]
        # store map
        errorMap = {}; warningMap = {}
        for row in pmeModel.table:
            errorMap[row[fullColTypeToColIdMap[NAME]]] =  row[fullColTypeToColIdMap[ERRORS]]
            warningMap[row[fullColTypeToColIdMap[NAME]]] =  row[fullColTypeToColIdMap[WARNINGS]]
        # Loop over each part
        if hasErrorsCol:
            errorsCol = self.currentColTypeToColIdMap[ERRORS]
        if hasWarningsCol:
            warningsCol = self.currentColTypeToColIdMap[WARNINGS]
        
        for krow in range(1, self.table.getNumRows()):
            partName = self.getItemValue(krow, NAME)
            if partName not in errorMap.keys(): continue
            self.setItemValue(krow, ERRORS, errorMap[partName])
            self.setItemValue(krow, WARNINGS, warningMap[partName])

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def isInstanceDependent(self, model, partName):
        a = model.rootAssembly
        part = model.parts[partName]
        for instanceName in a.instances.keys():
            instance = a.instances[instanceName]
            if not hasattr(instance, 'partName'): continue
            if instance.partName == partName:
                return instance.dependent

        # Instance does not exist, go ahead an mesh it
        return True
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def show(self):

        from kernelAccess import mdb, session

        model = mdb.models[self.form.modelName]
        model.materials.registerQuery(self.fillMaterialList, True)
        if hasattr(mdb.customData, 'pmeModels'):
            mdb.customData.pmeModels.registerQuery(self.updateTableState, True)
        AFXDataDialog.show(self)
                
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def hide(self):

        from kernelAccess import mdb, session

        model = mdb.models[self.form.modelName]
        model.materials.unregisterQuery(self.fillMaterialList)
        if hasattr(mdb.customData, 'pmeModels'):
            mdb.customData.pmeModels.unregisterQuery(self.updateTableState)
        AFXDataDialog.hide(self)