from abaqus import *
from abaqusConstants import *
from abq_Pme.pmeModel import PmeModel
from abq_Pme.pmeConstants import *
from abq_Pme.pmeUtils import getDefaultPartSettings

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def InitializeCustomData(modelName=None):
    if modelName == None:
        modelName = session.sessionState[session.currentViewportName]['modelName']
    if not hasattr(mdb.customData, 'pmeModels'):
        mdb.customData.Repository('pmeModels', PmeModel)
    if not mdb.customData.pmeModels.has_key(modelName):
        mdb.customData.PmeModel(modelName)
    
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def setPartInfo(modelName, partName=None):
    import customKernel
    session.customData.partInfo = {}
    model = mdb.models[modelName]
    part = model.parts[partName]
    if len(part.cells) == 0 and len(part.faces) == 0:
        return # Skip orphan mesh parts        
    session.customData.partInfo[partName] = getDefaultPartSettings(model, partName)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def hasVirtualTopology(part):
    for featureName in part.features.keys():
        if featureName[0:7] == "PME_Combine":
            return True
    return False

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def createSkinSurf(modelName, nextPartName, partName, faces):
    createPickedSurface(modelName, nextPartName, partName, faces, 'PME_OuterSkin')
    createPickedSet(modelName, nextPartName, partName, faces, 'PME_OuterSkin')

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def createBotSurf(modelName, nextPartName, partName, faces):
    createPickedSurface(modelName, nextPartName, partName, faces, 'PME_BotSurf')

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def createPickedSurface(modelName, nextPartName, partName, faces, surfName):
    model = mdb.models[modelName]
    p = model.parts[partName]
    surfFaces = None
    for face in faces:
        if surfFaces:
            surfFaces = surfFaces + p.faces[face.index:face.index+1]
        else:
            surfFaces = p.faces[face.index:face.index+1]
    p.Surface(name=surfName, side1Faces=surfFaces)
    if nextPartName != "None":
       p = model.parts[nextPartName]
       session.viewports[session.currentViewportName].setValues(displayedObject=p)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def createPickedSet(modelName, nextPartName, partName, faces, surfName):
    model = mdb.models[modelName]
    p = model.parts[partName]
    surfFaces = p.faces[0:0]
    for face in faces:
        surfFaces = surfFaces + p.faces[face.index:face.index+1]
    p.Set(name=surfName, faces=surfFaces)
    if nextPartName != "None":
       p = model.parts[nextPartName]
       session.viewports[session.currentViewportName].setValues(displayedObject=p)


