from abaqusGui import *
from abaqusConstants import *
from abq_Pme.pmeConstants import *

class PmeSurfaceProcedure(AFXProcedure):

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, owner):
        
        AFXProcedure.__init__(self, owner)

        self.cmd = AFXGuiCommand(self, '_UNSPECIFIED','abq_Pme.pmeModule', TRUE)
        self.partNameKw = AFXStringKeyword(self.cmd, 'partName', TRUE)
        self.modelNameKw = AFXStringKeyword(self.cmd, 'modelName', TRUE)
        self.nextPartNameKw = AFXStringKeyword(self.cmd, 'nextPartName', TRUE)
        self.facesKw = AFXObjectKeyword(self.cmd, 'faces', TRUE)
        self.step1 = None
        self.step2 = None
        self.partNames = []
        self.curPartNum = 0
        self.db = None
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setValues(self, db, modelName, partNames, method, surfaceName):
        self.db = db
        self.curPartNum = 0
        self.modelName = modelName
        self.partNames = partNames
        self.surfaceName = surfaceName
        self.cmd.setMethod(method)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def activate(self):
        self.db.hide()
        AFXProcedure.activate(self)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getFirstStep(self):
        
        curPartName = self.partNames[self.curPartNum]
        self.partNameKw.setValue(curPartName)
        if self.curPartNum < len(self.partNames)-1:
             self.nextPartNameKw.setValue(self.partNames[self.curPartNum+1])
        else:
            self.nextPartNameKw.setValue("None")
        title = "Select bottom surface of part '%s'." % curPartName
        self.step1 = AFXPickStep(self, self.facesKw, title, FACES, numberToPick=MANY,
                                 highlightLevel=1, sequenceStyle=TUPLE)
        
        from kernelAccess import mdb
        p = mdb.models[self.modelName].parts[curPartName]
        if p.surfaces.has_key(self.surfaceName):
            self.step1.addSurfaceSelection(self.surfaceName)
        return self.step1
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getNextStep(self, previousStep):
        if self.curPartNum != len(self.partNames)-1:
            curPartName = self.partNames[self.curPartNum]
            self.submitCommand(curPartName)
            self.curPartNum = self.curPartNum + 1
            curPartName = self.partNames[self.curPartNum]
            self.partNameKw.setValue(curPartName)
            if self.curPartNum == len(self.partNames)-1:
                self.nextPartNameKw.setValue("None")
            else:
                self.nextPartNameKw.setValue(self.partNames[self.curPartNum+1])
            title = "Select bottom surface of part '%s'." % curPartName
            self.step2 = AFXPickStep(self, self.facesKw, title, FACES, numberToPick=MANY,
                                     highlightLevel=1, sequenceStyle=TUPLE)
            return self.step2
        else:
            curPartName = self.partNames[self.curPartNum]
            self.nextPartNameKw.setValue("None")
            self.partNameKw.setValue(curPartName)
            self.submitCommand(curPartName)
            self.cmd.deactivate()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def submitCommand(self, curPartName, debug=None):
        # Replace first part name with current part name in cmdString
        self.modelNameKw.setValue(self.modelName)
        firstPartName =  self.partNames[0]
        cmdString = self.cmd.getCommandString()
        if debug:
            print "%s in submitCommand" % debug
            print "  firstPartName = ", firstPartName
            print "  curPartName = ", curPartName
            print "  cmdString = ", cmdString
        cmdString = cmdString.replace(firstPartName,curPartName)
        if debug:
            print "  newCmdString = ", cmdString
        self.sendCommandString(cmdString)
        
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onBackup(self):
        self.curPartNum = self.curPartNum-1
        curPartName = self.partNames[self.curPartNum]
        self.partNameKw.setValue(curPartName)
        self.nextPartNameKw.setValue(self.partNames[self.curPartNum+1])
        self.changePart(curPartName)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def changePart(self, partName):
        cmdString = "p=mdb.models['%s'].parts['%s']\n" % (self.modelName,partName)
        cmdString = cmdString + "session.viewports[session.currentViewportName].setValues(displayedObject=p)\n"
        self.sendCommandString(cmdString)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def handleException(self, exception):
        
        exceptionType = exception[0]
        exceptionValue = exception[1]

        # Do some special error handling here

        # Post an error dialog
        #
        showAFXErrorDialog(self.db, str(exceptionValue))
        self.db.show()
 
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onCancel(self):
        self.db.show()
        return None

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def doCustomTasks(self):
        self.db.show()
