from abaqusConstants import *
from abq_Pme.pmeConstants import *

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def getDefaultPartSettings(model, partName):
    defaultElemShape = "Hex"
    defaultSkin = False
    defaultRigid = False
    defaultMethod = "Direct"
    defaultMatName = None
    defaultOrder = "Linear"
    defaultFormulation = "Reduced Integration"
    if len(model.materials.keys()) > 0:
        matNames = model.materials.keys()
        defaultMatName = matNames[0]
    part = model.parts[partName]
    defaultSize = part.getPartSeeds(DEFAULT_SIZE)
    if len(part.cells) > 0:
        defaultThickness = " "
        technique = part.getMeshControl(region=part.cells[0], attribute=TECHNIQUE)
        if technique == SWEEP:
            defaultTechnique = "Solid Sweep"
        elif  technique == STRUCTURED:
            defaultTechnique = "Solid Structured"
        else:
            defaultTechnique = "Solid Free"
        if defaultTechnique in ("Solid Sweep","Solid Structured"):
            defaultElemShape = "Hex"
        else:
            defaultElemShape = "Tet"
            defaultFormulation = "Modified"
    elif len(part.faces) > 0:
        technique = part.getMeshControl(region=part.faces[0], attribute=TECHNIQUE)
        if technique == STRUCTURED:
            defaultTechnique = "Shell Structured"
        elif technique == SWEEP:
            defaultTechnique = "Shell Sweep"
        else:
            defaultTechnique = "Shell Free"
        defaultElemShape = "Quad Shell"
    else:
        defaultTechnique = None
    meshStatus = getMeshStatus(model, part)
    if defaultTechnique in ("Solid Sweep", "Solid Structured", "Solid Free"):
        defaultThickness = ''
    else:
        defaultThickness = 1.0
    volume = part.getVolume()
    if len(part.cells)==0:
        volume = 0.0
        for face in part.faces:
            volume = volume + face.getSize(printResults=False)
    volume = formatFloat(volume)
    partInfo = range(len(fullColTypeToColIdMap.keys()))
    partInfo[fullColTypeToColIdMap[STATUS]] = meshStatus
    partInfo[fullColTypeToColIdMap[NAME]] = partName
    partInfo[fullColTypeToColIdMap[TECHNIQUE]] = defaultTechnique
    partInfo[fullColTypeToColIdMap[ELEMSHAPE]] = defaultElemShape
    partInfo[fullColTypeToColIdMap[GENERATIONMETHOD]] = defaultMethod
    partInfo[fullColTypeToColIdMap[GEOMETRICORDER]] = defaultOrder
    partInfo[fullColTypeToColIdMap[FORMULATION]] = defaultFormulation
    partInfo[fullColTypeToColIdMap[ELEMSIZE]] = defaultSize
    partInfo[fullColTypeToColIdMap[MATERIAL]] = defaultMatName
    partInfo[fullColTypeToColIdMap[ERRORS]] = 0.0
    partInfo[fullColTypeToColIdMap[WARNINGS]] = 0.0
    partInfo[fullColTypeToColIdMap[RIGID]] = defaultRigid
    partInfo[fullColTypeToColIdMap[SKIN]] = defaultSkin
    partInfo[fullColTypeToColIdMap[THICKNESS]] = defaultThickness
    partInfo[fullColTypeToColIdMap[VOLUME]] = volume
    return partInfo

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def formatFloat(number):
    if number < 1.E-4:
        number = "%*.*E" % (10,3,number)
    elif number < 1.E6:
        number = "%*.*f" % (10,1,number)
    else:
        number = "%*.*E" % (10,3,number)
    return number.strip()
    
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def getMeshStatus(model,part):
    a = model.rootAssembly
    if len(part.elements) == 0:
        return ""
    else:
        return "Y"

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def getElemShape(model, part):
    mpName = part.name + partSuffix
    if mpName in model.parts.keys(): return CONT_SHELL
    if len(part.cells) > 0:
        shape = part.getMeshControl(part.cells[0], ELEM_SHAPE)
        if shape in (HEX,HEX_DOMINATED,WEDGE):
            if len(part.elements) > 0:
                elemType = part.elements[0].type
                if type in ('SC6R', 'SC8R'):
                    return CONT_SHELL
            return HEX
        else:
            return TET
    else:
        if len(part.faces) == 0:
            # Mesh Part or Orhpan Mesh (can't support)
            return None 
        else:
            return SHELL

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def formatMessage(msg, maxLineWidth=80, numIndentSpaces=6):
    line, remainingMessage = getLine(msg, maxLineWidth, numIndentSpaces)
    formattedMessage = line
    while len(remainingMessage) > 0:        
        line, remainingMessage = getLine(remainingMessage, maxLineWidth,
                                         numIndentSpaces)
        formattedMessage =  '%s%s' % (formattedMessage,line)
    return formattedMessage

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def getLine(msg, maxLineWidth=80, numIndentSpaces=6):
    lineWidth = maxLineWidth-numIndentSpaces
    if len(msg) <= lineWidth:
        line = "%s%s\n" % (numIndentSpaces*' ', msg)
        return line, ''
    for col in range(lineWidth,1,-1):
        if msg[col] == ' ':
            line = "%s%s\n" % (numIndentSpaces*' ', msg[0:col])
            remainingMessage = msg[col+1:]
            return line, remainingMessage
    line =  "%s%s\n" % (numIndentSpaces*' ', msg[0:maxLineWidth-numIndentSpaces])
    return line, msg[maxLineWidth-numIndentSpaces:]
