from abaqusGui import *
from abaqusConstants import *

###########################################################################
# Class definition
###########################################################################

class PmeVisibleColumnsDB(AFXDialog):

    numInstance = 0
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, mainDB, defaultColIdToColTypeMap, tableFeatures, currentColTypeToColIdMap):
    
        PmeVisibleColumnsDB.numInstance += 1

        AFXDialog.__init__(self, "Set Visible Columns", self.OK|self.CANCEL)

        #GUILOG
        getAFXAliasMap().setPrefix(self, 'PmeVisibleColumnsDB-%d' % PmeVisibleColumnsDB.numInstance)
        #ENDGUILOG

        self.mainDB = mainDB
        self.defaultColIdToColTypeMap = defaultColIdToColTypeMap
        self.currentColTypeToColIdMap = currentColTypeToColIdMap
        
        FXMAPFUNC(self, SEL_COMMAND, AFXDialog.ID_CLICKED_OK, PmeVisibleColumnsDB.onClick)
       
        # Set default button
        #
        vf = FXVerticalFrame(self, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X)
        self.checkButtons = []
        for kcol in range(len(defaultColIdToColTypeMap.keys())):
            colType = defaultColIdToColTypeMap[kcol]
            desc = tableFeatures[colType][0]
            desc = desc.replace("\n", " ")
            newDesc = desc[0]
            for k in range(1,len(desc)):
                if desc[k].isupper():
                    newDesc = newDesc + desc[k].swapcase()
                else:
                    newDesc = newDesc + desc[k]
            self.checkButtons.append(FXCheckButton(vf, newDesc))
            if currentColTypeToColIdMap.has_key(colType):
                self.checkButtons[-1].setCheck(True)

        i=1
        for chkButton in self.checkButtons:
            getAFXAliasMap().setName(chkButton, 'chkButton-%s' % i)
            i+=1
            
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def onClick(self, sender, sel, ptr):
        if SELID(sel) in (AFXDialog.ID_CLICKED_OK,):
            currentColIdToColTypeMap = {}
            visCol = 0
            for kcol in range(len(self.defaultColIdToColTypeMap.keys())):
                colType = self.defaultColIdToColTypeMap[kcol]
                if self.checkButtons[kcol].getCheck():
                    currentColIdToColTypeMap[visCol] = colType
                    visCol = visCol + 1
            if visCol == 0:
                msg = "At least one column must be visible."
                showAFXErrorDialog(getAFXApp().getAFXMainWindow(), msg)                
                return 0
            self.mainDB.changeColumnArrangement(currentColIdToColTypeMap)
            if SELID(sel) == AFXDialog.ID_CLICKED_OK:
                self.hide()
        return 1
